from pymongo import MongoClient
from bson.objectid import ObjectId
from datetime import datetime


class MergeSortDocs:
    def __init__(self):
        pass

    def _split(self, documents):
        documents_count = len(documents)
        if documents_count > 1:
            index = documents_count // 2
            return documents[:index], documents[index:]
        else:
            return documents, None

    def _merge(self, list_0, list_1):
        output_list = []
        index_0 = 0
        index_1 = 0
        while index_0 < len(list_0) and index_1 < len(list_1):
            date_0 = datetime.strptime(list_0[index_0]["dates_of_purchase"][-1], "%d-%m-%Y")
            date_1 = datetime.strptime(list_1[index_1]["dates_of_purchase"][-1], "%d-%m-%Y")
            if date_0 < date_1:
                output_list.append(list_1[index_1])
                index_1 += 1
            else:
                output_list.append(list_0[index_0])
                index_0 += 1
        return output_list + list_0[index_0:] + list_1[index_1:]

    def sort(self, documents):
        part_0, part_1 = self._split(documents)
        if part_1 is not None:
            return self._merge(self.sort(part_0), self.sort(part_1))
        else:
            return part_0


class Database:
    def __init__(self, mongodb):
        self._client = MongoClient(mongodb)
        self._db = self._client["autisto"]
        self._assets = self._db["assets"]
        self._decommissioned = self._db["decommissioned"]
        self.ss = None

    def get_assets(self, sort_by_latest=False):
        if sort_by_latest:
            assets = [document for document in self._assets.find({})]
            return MergeSortDocs().sort(assets)
        else:
            return self._assets.find({})

    def get_decommissioned(self):
        return self._decommissioned.find({})

    def _add_assets(self, order):
        if order.id is None:
            document = {
                "item_name": order.item_name,
                "category": order.category,
                "quantity": order.quantity,
                "life_expectancy_months": order.life_expectancy,
                "dates_of_purchase": [order.date.strftime("%d-%m-%Y") for _ in range(order.quantity)],
                "prices": [order.price for _ in range(order.quantity)],
                "decommissioned_id": None
            }
            self._assets.insert_one(document)
        else:
            document = self._assets.find_one({"_id": order.id})
            document.pop("_id")
            document["quantity"] += order.quantity
            index_to_append_at = 0
            for date in document["dates_of_purchase"]:
                if order.date < datetime.strptime(date, "%d-%m-%Y"):
                    break
                index_to_append_at += 1
            document["dates_of_purchase"] = [document["dates_of_purchase"][i] for i in range(index_to_append_at)] + \
                                            [order.date.strftime("%d-%m-%Y") for _ in range(order.quantity)] + \
                                            [document["dates_of_purchase"][i]
                                             for i in range(index_to_append_at, len(document["dates_of_purchase"]))]
            document["prices"] = [document["prices"][i] for i in range(index_to_append_at)] + \
                                 [order.price for _ in range(order.quantity)] + \
                                 [document["prices"][i] for i in range(index_to_append_at, len(document["prices"]))]
            self._assets.update_one({"_id": order.id}, {"$set": document})

    def _remove_assets(self, order):
        asset_document = self._assets.find_one({"_id": order.id})
        asset_document.pop("_id")
        if asset_document["decommissioned_id"] is None:
            decommissioned_document = self._assets.find_one({"_id": order.id})
            decommissioned_document.pop("decommissioned_id")
            decommissioned_document["quantity"] = order.quantity
            decommissioned_document["dates_of_purchase"] = asset_document["dates_of_purchase"][:order.quantity]
            decommissioned_document["prices"] = asset_document["prices"][:order.quantity]
            decommissioned_document.pop("_id")
            asset_document["decommissioned_id"] = self._decommissioned.insert_one(decommissioned_document).inserted_id
        else:
            decommissioned_document = self._decommissioned.find_one({"_id": asset_document["decommissioned_id"]})
            decommissioned_document["quantity"] += order.quantity
            decommissioned_document["dates_of_purchase"] += asset_document["dates_of_purchase"][:order.quantity]
            decommissioned_document["prices"] += asset_document["prices"][:order.quantity]
            decommissioned_document.pop("_id")
            self._decommissioned.update_one(
                {"_id": asset_document["decommissioned_id"]}, {"$set": decommissioned_document})
        asset_document["quantity"] -= order.quantity
        if asset_document["quantity"] <= 0:
            self._assets.delete_one({"_id": order.id})
        else:
            asset_document["dates_of_purchase"] = asset_document["dates_of_purchase"][order.quantity:]
            asset_document["prices"] = asset_document["prices"][order.quantity:]
            self._assets.update_one({"_id": order.id}, {"$set": asset_document})

    def get_id_object(self, identifier):
        object_id = ObjectId(identifier)
        if self._assets.find_one({"_id": object_id}) is None:
            raise ValueError
        else:
            return object_id

    def name_already_used(self, item_name):
        return self._assets.find_one({"item_name": item_name}) is not None

    def get_quantity(self, identifier):
        return self._assets.find_one({"_id": ObjectId(identifier)})["quantity"]

    def execute_orders(self, orders):
        for order in orders:
            if order.action == "add":
                self._add_assets(order)
            elif order.action == "remove":
                self._remove_assets(order)
            else:
                raise ValueError
        self.ss.console.clean_up(orders_only=True)



import json
import time
import pytest
import random
import string
import gspread
from pathlib import Path
from datetime import datetime
from autisto.spreadsheet import get_config, to_1_based, START_ROW, START_COL, CONSOLE_COL_NAMES, INVENTORY_COL_NAMES, \
    SPENDING_COL_NAMES

REFRESH_PERIOD = 60
ALPHABET = list(string.ascii_uppercase)
SHEET_NAMES = ["Console", "Inventory", "Spending"]


class Lock:
    def __init__(self):
        self._lock_path = Path("/tmp/autisto.lock")

    def acquire(self):
        while self._lock_path.exists():
            time.sleep(0.5)
        self._lock_path.touch()

    def release(self):
        self._lock_path.unlink()


lock = Lock()


@pytest.fixture
def spreadsheet():
    config = get_config()
    with open("client_credentials.json", "r") as client_credentials:
        gc = gspread.service_account_from_dict(json.load(client_credentials))
    name = f"Inventory {config['spreadsheet_uuid']}"
    return gc.open(name)


@pytest.mark.order(1)
def test_sheets_creation(spreadsheet):
    for sheet in SHEET_NAMES:
        _ = spreadsheet.worksheet(sheet)
    assert True


@pytest.mark.order(3)
def test_column_titling(spreadsheet):
    sheets_to_titles = {"Console": CONSOLE_COL_NAMES, "Inventory": INVENTORY_COL_NAMES, "Spending": SPENDING_COL_NAMES}
    lock.acquire()
    for sheet, titles in sheets_to_titles.items():
        start_row = to_1_based(START_ROW) + 1 if sheet == "Inventory" else to_1_based(START_ROW)
        row_values = spreadsheet.worksheet(sheet).row_values(start_row)[START_COL:]
        for i, col_name in enumerate(titles):
            assert row_values[i] == col_name
    lock.release()


@pytest.mark.order(2)
def test_sheets_maintaining(spreadsheet):
    class RandomCoordinates:
        def __init__(self):
            self.row = random.randint(1, 1000)
            self.col = random.randint(1, len(ALPHABET))

    litter = ["She's got the wings and teeth of a African bat",
              "Her middle name is Mudbone and on top of all that",
              "Your mama got a glass eye with the fish in it"]
    cells_to_litter = {sheet: [] for sheet in SHEET_NAMES}
    for sheet in cells_to_litter.keys():
        for _ in range(len(litter)):
            cells_to_litter[sheet].append(RandomCoordinates())
    for sheet in SHEET_NAMES:
        worksheet = spreadsheet.worksheet(sheet)
        for i in range(len(litter)):
            worksheet.update_cell(
                cells_to_litter[sheet][i].row,
                cells_to_litter[sheet][i].col,
                litter[i]
            )
    time.sleep(REFRESH_PERIOD + 10)

    for sheet in SHEET_NAMES:
        worksheet = spreadsheet.worksheet(sheet)
        for i in range(len(litter)):
            cell_coordinates = cells_to_litter[sheet][i]
            if sheet == "Console":
                if cell_coordinates.row == 1:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
                elif cell_coordinates.col == 1 or to_1_based(START_COL) + len(
                        CONSOLE_COL_NAMES[:-2]) <= cell_coordinates.col:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
                elif cell_coordinates.row != 2:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value == litter[i]
            elif sheet == "Inventory":
                if cell_coordinates.row == 1 or to_1_based(START_ROW) + 2 <= cell_coordinates.row:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
                elif cell_coordinates.col == 1 or \
                        to_1_based(START_COL) + len(INVENTORY_COL_NAMES) <= cell_coordinates.col:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
            elif sheet == "Spending":
                if cell_coordinates.row == 1 or to_1_based(START_ROW) + 1 <= cell_coordinates.row:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
                elif cell_coordinates.col == 1 or \
                        to_1_based(START_COL) + len(SPENDING_COL_NAMES) <= cell_coordinates.col:
                    assert worksheet.cell(cell_coordinates.row, cell_coordinates.col).value is None
            else:
                assert False


def add_remove_item(console, item_data, action_token):
    row_values = []
    for col_name in CONSOLE_COL_NAMES:
        if col_name == "Action <ADD/REMOVE>":
            row_values.append(action_token)
        elif col_name == "Done? <Y>":
            row_values.append("y")
        else:
            try:
                row_values.append(item_data[col_name])
            except KeyError:
                row_values.append("")
    row = random.randint(3, 1000)
    console.update(f"{ALPHABET[START_COL]}{row}:{ALPHABET[START_COL + len(CONSOLE_COL_NAMES) - 1]}{row}", [row_values])


example_purchase_0 = {
    "Quantity": "2",
    "Date of purchase [DD-MM-YYYY]": datetime.now().strftime("%d-%m-%Y"),
    "Unit price [PLN]": "3189,99",
    "Item name": "Glock 26, 9mm",
    "Category": "Self-defense only",
    "Life expectancy [months]": "300"
}


@pytest.mark.order(4)
def test_adding(spreadsheet):
    console = spreadsheet.worksheet("Console")
    add_remove_item(console, example_purchase_0, "a")
    time.sleep(REFRESH_PERIOD + 10)

    inventory = spreadsheet.worksheet("Inventory")
    total_value = int(example_purchase_0["Quantity"]) * float(example_purchase_0["Unit price [PLN]"].replace(",", "."))
    assert total_value == float(inventory.cell(
        to_1_based(START_ROW), to_1_based(START_COL) + INVENTORY_COL_NAMES.index("Total value [PLN]")).value)
    row_values = inventory.row_values(to_1_based(START_ROW) + 2)
    for i, col_name in enumerate(INVENTORY_COL_NAMES):
        if col_name in ["Category", "Item name", "Quantity", "Life expectancy [months]"]:
            assert example_purchase_0[col_name] == row_values[i + START_ROW]
        elif col_name == "Average unit value [PLN]":
            assert float(example_purchase_0["Unit price [PLN]"].replace(",", ".")) == float(row_values[i + START_ROW])
        elif col_name == "Depreciation [PLN]":
            assert 0. == float(row_values[i + START_ROW])


example_purchase_1 = {
    "ID": None,
    "Quantity": "1",
    "Date of purchase [DD-MM-YYYY]": "28-12-1996",  # nice date
    "Unit price [PLN]": "1570,10"
}


@pytest.mark.order(5)
def test_appending(spreadsheet):
    console = spreadsheet.worksheet("Console")
    inventory = spreadsheet.worksheet("Inventory")
    example_purchase_1["ID"] = inventory.cell(to_1_based(START_ROW) + 2, to_1_based(START_COL)).value
    add_remove_item(console, example_purchase_1, "ADD")
    time.sleep(REFRESH_PERIOD + 10)

    assert "3" == inventory.cell(
        to_1_based(START_ROW) + 2, to_1_based(START_COL) + INVENTORY_COL_NAMES.index("Quantity")).value
    assert 0. < float(inventory.cell(
        to_1_based(START_ROW) + 2, to_1_based(START_COL) + INVENTORY_COL_NAMES.index("Depreciation [PLN]")).value)


@pytest.mark.order(6)
def test_removing(spreadsheet):
    console = spreadsheet.worksheet("Console")
    inventory = spreadsheet.worksheet("Inventory")
    item_data = {
        "ID": example_purchase_1["ID"],
        "Quantity": "2"
    }
    add_remove_item(console, item_data, "r")

    time.sleep(REFRESH_PERIOD + 10)
    assert "1" == inventory.cell(
        to_1_based(START_ROW) + 2, to_1_based(START_COL) + INVENTORY_COL_NAMES.index("Quantity")).value
    total_value = float(example_purchase_0["Unit price [PLN]"].replace(",", "."))
    assert total_value == float(inventory.cell(
        to_1_based(START_ROW), to_1_based(START_COL) + INVENTORY_COL_NAMES.index("Total value [PLN]")).value)


@pytest.mark.order(7)
def test_spending(spreadsheet):
    spending = spreadsheet.worksheet("Spending")
    lock.acquire()
    now = datetime.now()

    date = datetime.strptime(example_purchase_0["Date of purchase [DD-MM-YYYY]"], "%d-%m-%Y")
    offset = (now.year - date.year) * 12 + now.month - date.month
    total_price = int(example_purchase_0["Quantity"]) * float(example_purchase_0["Unit price [PLN]"].replace(",", "."))
    assert total_price == float(spending.cell(to_1_based(START_ROW) + 1 + offset, to_1_based(START_COL) + 2).value)

    date = datetime.strptime(example_purchase_1["Date of purchase [DD-MM-YYYY]"], "%d-%m-%Y")
    offset = (now.year - date.year) * 12 + now.month - date.month
    total_price = int(example_purchase_1["Quantity"]) * float(example_purchase_1["Unit price [PLN]"].replace(",", "."))
    assert total_price == float(spending.cell(to_1_based(START_ROW) + 1 + offset, to_1_based(START_COL) + 2).value)
    lock.release()



import csv
import urllib.request
from datetime import datetime
from dateutil.relativedelta import relativedelta

URL = "https://stat.gov.pl/download/gfx/portalinformacyjny/pl/defaultstronaopisowa/4741/1/1/miesieczne_wskazniki_cen_" \
      "towarow_i_uslug_konsumpcyjnych_od_1982_roku.csv"


class FinanceModule:
    def __init__(self):
        self._calc_accumulated_inflation(self._extract_month_over_month_inflation(self._load_inflation_data()))

    def _load_inflation_data(self):
        response = urllib.request.urlopen(URL)
        return csv.reader([line.decode("iso-8859-2") for line in response.readlines()], delimiter=";")

    def _extract_month_over_month_inflation(self, raw_inflation_data):
        month_over_month_inflation_data = {}
        for row in raw_inflation_data:
            if "Poprzedni" in row[2]:
                try:
                    inflation_rate = float(row[5].replace(",", ".")) / 100
                except ValueError:
                    continue
                if row[3] in month_over_month_inflation_data.keys():
                    month_over_month_inflation_data[row[3]][row[4]] = inflation_rate
                else:
                    month_over_month_inflation_data[row[3]] = {row[4]: float(row[5].replace(",", ".")) / 100}
        return month_over_month_inflation_data

    def _calc_accumulated_inflation(self, month_over_month_inflation_data):
        self._current_time = datetime.now()
        self._accumulated_inflation = {}
        accumulated_inflation = 1.
        for year in reversed(range(1982, self._current_time.year + 1)):
            self._accumulated_inflation[str(year)] = {}
            for month in reversed(range(1, 13)):
                try:
                    accumulated_inflation *= month_over_month_inflation_data[str(year)][str(month)]
                except KeyError:
                    pass
                self._accumulated_inflation[str(year)][str(month)] = accumulated_inflation

    def calc(self, document):
        total_value = 0.
        depreciation = 0.
        for i in range(document["quantity"]):
            purchase_date = datetime.strptime(document["dates_of_purchase"][i], "%d-%m-%Y")
            try:
                adjusted_value = document["prices"][i] * self._accumulated_inflation[
                    str(purchase_date.year)][str(purchase_date.month)]
            except KeyError:
                adjusted_value = document["prices"][i] * self._accumulated_inflation["1983"]["1"]
            total_value += adjusted_value
            relative_delta = relativedelta(self._current_time, purchase_date)
            months_passed = relative_delta.years * 12 + relative_delta.months
            depreciation_ratio = min(1., months_passed / document["life_expectancy_months"])
            depreciation += adjusted_value * depreciation_ratio
        return total_value, depreciation



import os
import setuptools

VERSION = "1.0"


def load_readme():
    with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
        readme = f.read()
    return readme


def load_requirements(file_name):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), file_name)) as f:
        requirements = f.readlines()
    return [req.replace("\n", "") for req in requirements]


if __name__ == "__main__":
    entry_points = {
        "console_scripts": [
            "autisto = autisto.utils:check_setup"
        ]
    }

    python_requires = '>=3.10'
    install_requires = load_requirements("requirements.txt")

    packages = [
        "autisto"
    ]

    project_urls = {
        "GitHub": "https://github.com/jan-grzybek/autisto"
    }

    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3 :: Only"
    ]

    setuptools.setup(
        name="autisto",
        version=VERSION,
        author="Jan Grzybek",
        author_email="lyre_embassy_0n@icloud.com",
        description="Basic accounting (?) program integrated with Google Sheets to fulfill my own autistic needs of "
                    "tracking the shit owned.",
        long_description=load_readme(),
        long_description_content_type="text/markdown",
        entry_points=entry_points,
        license="MIT",  # some serious research ...
        install_requires=install_requires,
        python_requires=python_requires,
        packages=packages,
        classifiers=classifiers,
        project_urls=project_urls
    )



import os
import sys
import json
import uuid
import pymongo
from autisto.daemons import get_platform
from autisto.database import Database

CONFIG_DIR = "~/.config/autisto/"
CONFIG_FILE_NAME = "config.json"


def to_1_based(index):
    return index + 1


class FaultyOrder(Exception):
    pass


def check_for_positive_int(value):
    value = float(value)
    if value % 1 != 0 or value < 1:
        raise ValueError
    return int(value)


def check_for_positive_float(value):
    value = float(value)
    if value < 0.:
        raise ValueError
    return value


class Order:
    def __init__(self, row, action, identifier, quantity,
                 date=None, price=None, item_name=None, category=None, life_expectancy=None):
        self.action = action
        self.id = identifier
        self.quantity = quantity
        self.date = date
        self.price = price
        self.item_name = item_name
        self.category = category
        self.life_expectancy = life_expectancy
        self.row = row


def get_config():
    with open(os.path.expanduser(os.path.join(CONFIG_DIR, CONFIG_FILE_NAME)), "r") as config_file:
        config = json.load(config_file)
    if "refresh_period" not in config.keys():
        config["refresh_period"] = 900
        with open(os.path.expanduser(os.path.join(CONFIG_DIR, CONFIG_FILE_NAME)), "w") as config_file:
            config_file.write(json.dumps(config))
    return config


def do_config():
    print("Attempting connection to mongoDB ...")
    try:
        for _ in Database("mongodb://localhost:27017/").get_assets():
            break
        print("Success.")
    except pymongo.errors.ServerSelectionTimeoutError as e:
        print(e)
        print("Is mongoDB installed on the system?")
        sys.exit(1)
    print("\nChecking for root privileges ...")
    if os.geteuid() == 0:
        print("Success.")
    else:
        print("Please rerun with sudo.")
        print("   sudo autisto")
        sys.exit(1)
    print("\nHello. Looks like Autisto personal accountant has not been set up yet.")
    print("Have you already set up a Google Service Account? If not, please first follow instructions here: "
          "https://docs.gspread.org/en/latest/oauth2.html")
    print("\nPlease provide path to the .json file with Service Account credentials.")
    while True:
        path = input("Path to credentials: ")
        try:
            credentials = json.load(open(os.path.expanduser(path), "r"))
            break
        except FileNotFoundError as e:
            print(e)
    print("\nPlease provide email address that you want to access spreadsheet with.")
    email = input("Email address: ")
    print("\nWhat should be the refresh period for the spreadsheet?")
    refresh_period = None
    while refresh_period is None:
        try:
            refresh_period = int(input("Refresh period in seconds [recommended: 900]: "))
        except ValueError:
            continue
    config = {
        "user_email": email,
        "spreadsheet_uuid": str(uuid.uuid4()),
        "refresh_period": refresh_period,
        "credentials": credentials
    }
    os.makedirs(os.path.expanduser(CONFIG_DIR), exist_ok=True)
    with open(os.path.expanduser(os.path.join(CONFIG_DIR, CONFIG_FILE_NAME)), "w") as config_file:
        config_file.write(json.dumps(config))
    print(f"\nThank you. Your config has been saved under {CONFIG_DIR}")
    print("\nSetting system daemon ...")
    get_platform().set_service()
    print("All done.")
    sys.exit(0)


def check_setup():
    try:
        get_config()
        get_platform().service_active()
    except FileNotFoundError:
        do_config()
